
#include "ipp.h"
#include "adapt.h" 


int max_adapt(int max_size, int *max_time, int *result_build)  
{  
  int result=0;  
     
  for ( ; *max_time <= max_size && !result ; (*max_time)++ ) 
  {   

    if ( display_info > 2 ) 
       printf("\n SEARCH max_time: %d max_size %d \n", *max_time, max_size);  
 
   if ( *result_build && ( result = search_plan( *max_time ) ) != FALSE )   
      break;   
      /* couldn't find plan; check for unsolvability */   
      
    if ( same_as_prev_flag ) 
    {   
   
      /* start after graph has leveled off */   
      if ( previous_Scount == current_Scount ) 
      {   
	/* nothing has changed in last step -> unsolvable */   
	break;   
      } 
      else 
      {   
	/* still improvements going on */   
	previous_Scount = current_Scount;   
      }       
    }   
    /* we must continue search with one more layer */          
           
    if( adapt.curr_plan_length <= *max_time )   
    {     
	build_graph_layer(FALSE);   
	*result_build= are_there_non_exclusive( goal_facts, *max_time);
	if(*result_build==FALSE)
           result= *result_build;
           
	adapt.curr_plan_length++;   
    }     
  }   
  return (result);  
}  



int adapt_build_graph(int *max_time, int max_size)   
{   
  struct tms build_s, build_e; 
  int reached_goals, time=0, first=0, minim;        
  same_as_prev_flag=0;        
  
  times(&build_s); 
  
  if (adapt.change==CHANGE_GOAL)   
  {   
       
      for( minim=1; minim<max_size; minim++ )   
      {   
	  if(minim>=adapt.curr_plan_length)   
	  {   
	      build_graph_layer(FALSE);   
	      adapt.curr_plan_length++;   
	  }   
	  
	  reached_goals = are_there_non_exclusive( goal_facts, minim );   
	  if ( reached_goals ) break;   
	  if ( same_as_prev_flag ) break;   
       }   
       
      *max_time=minim;    
        
  }  
  
  else 
  { 
      adapt.curr_plan_length=0;   
      if(fact_table)  
       	free_table(fact_table, MAX_PLAN+1);   
      fact_table=NULL;   
      if(op_table)
        free_table(op_table, MAX_PLAN);   
      op_table=NULL;   

      time=1;   
      /* build graph until non-exclusive goals are reached */   
      reached_goals = build_graph( &time, FALSE );   
   
      /* plan until goals are reached or graph has leveled off or time>max_size */   
      for( ; time <=max_size ; time++ ) 
      {   
	reached_goals = are_there_non_exclusive( goal_facts, time );   
	if ( reached_goals ) 
	{   
           if ( display_info && time > 0 && first )           
              printf("\nADJ: goals first reachable in %d time steps \n", time);
          
  	   break;   
	}   
	if ( same_as_prev_flag ) break;   
	build_graph_layer(FALSE);   
      }   
       
      adapt.curr_plan_length=time;   
      *max_time = time; 
    } 
 
  times(&build_e); 
 
  build_ad_time += ( float ) ( ( build_e.tms_utime - build_s.tms_utime + 
		        	 build_e.tms_stime - build_s.tms_stime  ) / 100.0 ) ;
 
  return (reached_goals);   
}    



// Detect if initial facts or goals are changed, or both

int detect_plan_changes(int curr_init_time, int curr_max_time, int max_size, int *max_time)
{

   int result_build;

   if(adapt.change==CHANGE_GOAL || adapt.change==TRUE)    
   {  
      goal_facts=goal_table_vect[curr_max_time];   
	
      if (adapt.change==CHANGE_GOAL )
	 result_build=adapt_build_graph(max_time, max_size);			      	
   }   
       
   if(adapt.change==CHANGE_INITIAL || adapt.change==TRUE)    
   { 
      initial_facts=fact_table_vect[curr_init_time];   
      result_build=adapt_build_graph(max_time, max_size);

      de_alloc(*max_time);
   }

   return (result_build);
}



// Set in the new plan if initial facts or goals are changed, or both

void set_plan_changes(int *curr_init_time, int *curr_max_time)
{

   // we try to expand the planning graph backward, better because no new levels needed  ...

   if( ((adapt.alterna)%2) <1)
   {
      if (*curr_max_time < adapt.plan_length)
      {
	 (*curr_max_time)++;
	 adapt.change=CHANGE_GOAL;
      }
      else
	 if (*curr_init_time>0)
	 {
            (*curr_init_time)--;   
	    adapt.change=CHANGE_INITIAL;   
	 }
   }

   else
   {
      if (*curr_init_time>0)   
      {  
	 (*curr_init_time)--;   
	 adapt.change=CHANGE_INITIAL;   
      }   
      else   
	 if (*curr_max_time < adapt.plan_length)   
	 { 
	    (*curr_max_time)++;     
	    adapt.change=CHANGE_GOAL;   
	 }
   }

   adapt.alterna++;
}



// Action vector is updated, correction vector is updated too (up or down)

void update_act_vect(int curr_init_time, int curr_max_time, int max_time)
{
   int diff, i;
   vertex_list op;	

   diff=curr_max_time-curr_init_time;

   if(diff>max_time)
   {
      for(i=diff-1; i>=max_time; i--)
         down_act_vect(curr_init_time+i);
   }
	
   for(i=0; i<max_time; i++)
   {
      if(i>=diff  && diff < max_time )
	up_act_vect(curr_init_time+i);
      else
        free_act_vect(curr_init_time+i);
   }

   for(i=max_time-1; i>=0; i--)
   {
      get_next( op_table[i], INIT );/* initialise */
      while ( ( op = get_next( op_table[i], EXEC ) ) != NULL )
	 if(op->is_used && !op->is_noop)
            insert_act_vect(op->name, curr_init_time+i);	
   }

}



// Set current initial time , current max time and inconsistence time

void set_current_times(int *curr_init_time, int *curr_max_time, int *inc_time, int try)
{
   int i;

   for(i=0; i<adapt.plan_length && correct_vect[i]==0; i++);
   (*curr_init_time)=i;

   for(; i<adapt.plan_length && correct_vect[i]==try; i++);
   (*curr_max_time)=i;

   *inc_time = *curr_init_time ;

   if ( (*curr_max_time)==adapt.plan_length && (*curr_init_time)!=0 )
   {
      (*curr_init_time)--;
      // we couldn't search after, so we search before
      *inc_time = *curr_max_time;
   }
   else
      if( (*curr_init_time)==0 && (*curr_max_time)!=adapt.plan_length )
         (*curr_max_time)++;

}


// Correct vector is used during optimization process

void init_correct_vect(int curr_init_time, int max_time, int try)
{
  int i;

  for(i=0; i<=adapt.plan_length; i++)
     if ( (correct_vect[i]>0) && (correct_vect[i] < try) )
        correct_vect[i]=try;

  for(i=0; i<max_time; i++)
     correct_vect[curr_init_time+i]=try;
}



void update_correct_vect(int curr_init_time, int curr_max_time, int try)
{
  int i;

  for(i=curr_init_time; i<curr_max_time; i++)
        correct_vect[i]=try;
}



/* the setting up function  called by main   
 * simply sets up search and does the initial call   
 * as well as memoizing the result if search was a failure   
 */            
   
BOOLEAN adapt_plan(int curr_init_time)
{

  int max_size=MAX_SIZE;
  int result=FALSE, result_build;
  int max_time=1, curr_max_time;
  int inc_time=curr_init_time;

  expand(adapt.plan_length+20);    /* AAAAAAAAAAAAAAAAA */

  if (curr_init_time<adapt.plan_length)
     curr_max_time=curr_init_time+1;

  else  /*  if(curr_init_time>0) */
  {
     curr_max_time=curr_init_time;
     curr_init_time--;
  }

  same_as_prev_flag=0;
  adapt.change=TRUE;

  // we search for a plan of "max_size" maximum length starting from level "curr_init_time"

  while( (result==FALSE || result==PARTIAL_TIMEOUT) && adapt.curr_plan_length < MAX_PLAN )
  {

     adapt.try++;

     result_build=detect_plan_changes(curr_init_time, curr_max_time, max_size, &max_time);

     printf("\n Replanning Window <%d, %d>", curr_init_time, curr_max_time);
     printf("  (contraint violation of time %d)", inc_time);
     printf("  (iteration: %d  try: %d  max_size: %d)\n", adapt.iteraz, adapt.try, max_size);

     adapt.change=FALSE;

     result=max_adapt(max_size, &max_time, &result_build);

     if (result == TRUE)
     {
        printf("\n\nPLAN WINDOW TO ADAPT\n");
        print_facts_and_goals(curr_init_time, curr_max_time);

        printf("\nPARTIAL SOLUTION (time %d)\n",curr_init_time);
        print_partial_plan(max_time);
        de_alloc(max_time);

        update_act_vect(curr_init_time, curr_max_time, max_time);
        init_correct_vect(curr_init_time, max_time, adapt.try);

        define_facts(curr_init_time, adapt.plan_length);
     }

     else if (result == TIMEOUT)
	break;

     else
     {
        set_plan_changes(&curr_init_time, &curr_max_time);
        max_size+=adapt.step;
     }
  }

  return (result);
}



// it compute plan optimization (more compact solutions)

int resize_plan(char* fact_file_name)
{

  int max_size, result, result_build;
  int max_time=1, curr_init_time, curr_max_time;
  int inc_time;
  int max_size_count=0;
  BOOLEAN stop=FALSE;

  float plan_time;
  struct tms end_time;

  expand(adapt.plan_length+20); /* AAAAAAAAAAAAAAAAA */
  printf("\nCOMPUTING ADDITIONAL MORE COMPACT SOLUTIONS\n");

  while ( !stop )
  {

      adapt.iteraz++;

      set_current_times(&curr_init_time, &curr_max_time, &inc_time, adapt.try);

      same_as_prev_flag=0;
      adapt.curr_plan_length=0;
      adapt.change=TRUE;
      result= FALSE;

      // we search for a (better) plan of "max_size" maximum length starting from level "curr_init_time"

      while ( (result==FALSE || result==PARTIAL_TIMEOUT) && adapt.curr_plan_length < MAX_PLAN && !stop )
      {

        adapt.try++;

	max_size= curr_max_time-curr_init_time-1;

        result_build=detect_plan_changes(curr_init_time, curr_max_time, max_size, &max_time);

	adapt.change=FALSE;
        printf("\n Replanning Window <%d, %d>",curr_init_time,curr_max_time);
        printf("  (iteration: %d  try: %d  max_size: %d)\n", adapt.iteraz, adapt.try, max_size);

        result=max_adapt(max_size, &max_time, &result_build);

        times(&end_time);

	plan_time = ( float ) ( ( end_time.tms_utime - start_time.tms_utime +
 		                  end_time.tms_stime - start_time.tms_stime ) / 100.0 ) ;

	if ( result==TRUE )
	{

	   printf("\n\nPLAN WINDOW TO OPTIMIZE (%d)\n",adapt.iteraz);
           print_facts_and_goals(curr_init_time, curr_max_time);

	   if(display_info>1)
	   {
    	      printf("\nPARTIAL SOLUTION\n");
	      print_partial_plan(max_time);
	   }

  	   de_alloc(max_time);

  	   update_act_vect(curr_init_time, curr_max_time, max_time);
           update_correct_vect(curr_init_time, curr_max_time, adapt.try);

	   define_facts(curr_init_time, adapt.plan_length);

           end_control();

	
 	   if ( adapt.plan_length < adapt.best_plan_length ||
              ( adapt.plan_length==adapt.best_plan_length && adapt.num_actions < adapt.best_num_actions ))
           {

              printf("\nOPTIMIZED PLAN (%d)\n", adapt.iteraz);
              print_plan( adapt.curr_plan_length );
              print_adj_report(plan_time);
              store_adapted_plan( adapt.plan_length, fact_file_name, plan_time );
              fflush(stdout);

              // this is only for testing
              #ifdef TEST_ADJ
                 printf("\nAABBCC%d\t ADJ: spent \t %.2f  seconds total time; num. of levels \t %d num. of actions \t %d\n",
                        adapt.iteraz, plan_time, adapt.plan_length,  adapt.num_actions );

//                 exit(0);
              #endif /* TEST_ADJ */

 	      adapt.best_plan_length = adapt.plan_length;
              adapt.best_num_actions = adapt.num_actions;
 	   }
	
  	   else
	   {
	      printf("\nADJ: the previous plan is already optimized ! \n\n");
              stop=TRUE;
	   }	
	
	}
		
	else if ( result==TIMEOUT )
        {
           printf("\nADJ: timeout reached, please increase time limit! \n");
	   return(result);
	}	
	
	else
	   set_plan_changes(&curr_init_time, &curr_max_time);


        if ((curr_init_time<=0) && (curr_max_time>=adapt.plan_length))
        {
	   max_size_count++;	// iteration on the last replanning window
	
  	   if ( max_size_count>=2 )
 	      stop=TRUE;
        }
 	
 	if (stop)
 	{
           printf("\n\nBEST PLAN\n"); 	
           print_plan( adapt.curr_plan_length );
           print_adj_report(plan_time);  	

           // this is only for testing
           #ifdef TEST_ADJ
              printf("\nAABBCCO\t ADJ: spent \t %.2f  seconds total time; num. of levels \t %d num. of actions \t %d\n",
                     plan_time, adapt.plan_length,  adapt.num_actions );
           #endif /* TEST_ADJ */
        }
              			
      }

  }   
    
  return (result);
}  



void find_inc(int time_in, int time_end, int control)   
{   
  int i;   
  vertex_list ft, op;   
     
  adapt.num_inc=0;  
  for(i=0; i<= time_end; i++)  
    incons_vect[i]=0;
  
  
  for(i=time_in; i<=time_end; i++)
  {   
    get_next( fact_table[i], INIT );/* initialise */   
    while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL ) 
    {   
      if(( ft->is_goal && ft->is_true==FALSE) || (control && ft->is_true && me_fct(ft, i)))
      {   
	insert_incons(i-time_in);   
	printf("\nADJ: constraint violation '%s' at level %d \n", ft->name, i - time_in);
	break;   
      }   
      /* inserire controllo ME */  
    }   
        
    if(ft) continue;   
    if(i==time_end) break;   
       
    get_next( op_table[i], INIT );/* initialise */   
    while ( ( op = get_next( op_table[i], EXEC ) ) != NULL ) 
    {   
      if( op->is_used && me_op(op, i) )
      {   
	insert_incons(i-time_in);   
        printf("\n inconsistence %s level %d", op->name, i - time_in);
	break;  
      }   
    }   

  }    
   
  if( !adapt.num_inc )  
  { 
    if(display_info>2) 
        printf("\nNO INCONSISTENCE FOUND !");
  }  
  else 
     printf("\nADJ: total number of contraint violation in the plan = %d", adapt.num_inc);
   
  printf( "\n" );  
  
#ifdef TEST
  printf( "Incons:" );  
  for ( i=0; i<=time_end; i++ )   
    printf(" %d ", incons_vect[i]);
    
  printf( "\n" );  
#endif /* TEST */
}



/* Cerca le inconsistenze presenti nel planning graph corrente (di
 * lungh tot_time) ed aggiorna il vettore delle inconsistenze.
 * Il livello 0 del current planning graph corrisponde al livello time_glob
 */

void find_local_inc(int time_glob, int tot_time)
{
  int i;
  vertex_list ft, op;

  for(i=0; i<= tot_time; i++){
    get_next( fact_table[i], INIT );/* initialise */
    while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL )
    {
       if(( ft->is_goal && ft->is_true==FALSE) || ( ft->is_true && me_fct(ft, i)))
       {
	  insert_incons(i+time_glob);
  	  printf("\nADJ: fact inconsistence '%s' at level %d \n", ft->name, i + time_glob);
	  break;
       }
       /* inserire controllo ME */
    }

    if(ft) continue;
    if(i==tot_time) break;
    /* Il livello finale non contiene azioni */

    get_next( op_table[i], INIT );/* initialise */
    while ( ( op = get_next( op_table[i], EXEC ) ) != NULL )
    {
      if( op->is_used && me_op(op, i) )
      {
	insert_incons(i+time_glob);
        printf("\nADJ: action inconsistence '%s' at level %d \n", op->name, i + time_glob);
	break;
      }
    }


  }

#ifdef TEST
  if( !adapt.num_inc )
    printf("\nNO INCONSISTENCE FOUND ");
  else printf("\n FOUND  %d inconsistences", adapt.num_inc);
  printf( "\n" );

  printf( "Incons:" );
  for ( i=0; i<=adapt.plan_length; i++ )
    printf(" %d ", incons_vect[i]);

  printf( "\n" );
#endif /* TEST */

}



/* Make temporal projection from time step init_time
 * The global  planning  graph has length tot_time
 */

void define_facts(int init_time, int tot_time)
{
  int p_length, i, j, pos=init_time, time;
  vertex_list ft, op;
  action *temp_act;
  token_list temp;

  adapt.num_inc=0;
  for(i=0; i<= tot_time; i++)
    incons_vect[init_time+i]=0;

  initial_facts=fact_table_vect[init_time];
  goal_facts=store_goal_facts;

#ifdef TEST
  printf("\nPlan length %d \n Fact level %d:   ", adapt.plan_length, init_time);
  print_facts_and_goals(curr_init_time, curr_max_time);
#endif /*  TEST */

  expand(adapt.plan_length+10);
  de_alloc(adapt.plan_length);

  if(fact_table)
    free_table(fact_table, MAX_PLAN+1);
  fact_table=NULL;
  if(op_table)
    free_table(op_table, MAX_PLAN);
  op_table=NULL;

  p_length=1;
  /* build graph until non-exclusive goals are reached */
  build_graph( &p_length, FALSE );
  for ( temp = initial_facts; temp; temp = temp->next )
  {
    ft = lookup_from_table( fact_table[0], temp->item );
    if ( ft )
      ft->is_true = 1;

  }

  for(pos=init_time, p_length=0 , j=0; (pos+j)< tot_time;  p_length++, j++)
  {
    if(act_vect[pos+j])
    {
      time=act_vect[pos+j]->iPasso;
	
      for(temp_act=act_vect[pos+j]; temp_act; temp_act=temp_act->psaNext)
      {
	if(time!=temp_act->iPasso)
	  {
	    ins_noop_layer(p_length);
	    time=temp_act->iPasso;
	
	    p_length++;
	 	
	    printf("\nINSERISCO LIVELLO");
	    build_graph_layer(FALSE);
	  }
	
	op = lookup_from_table( op_table[p_length], temp_act->cStr );
	if (op)
	{
	    insert_operators(op, p_length );	
#ifdef TEST
	    printf("\nAction %s level %d ",op->name, pos+j);
#endif
	}
	
	else
	{
#ifdef TEST
	  printf("\n There are some errors in setup; debug me please.\n");
	  printf("\n Action %s, level %d\n",temp_act->cStr, pos+j);
#endif	
	  insert_incons(pos+j);
	}
      }

    }
    ins_noop_layer(p_length);
    if( fact_table_vect[pos+j+1] )
    {
      free_token_list(fact_table_vect[pos+j+1]);
      fact_table_vect[pos+j+1]=NULL;

    }
    get_next( fact_table[p_length+1], INIT );/* initialise */
    while ( ( ft = get_next( fact_table[p_length+1], EXEC ) ) != NULL )
    {
      if( ft->is_true )
      {
        temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );
        temp->item = new_token( strlen( ft->name ) + 1 );
        strcpy(temp->item, ft->name);
        temp->next = fact_table_vect[pos+j+1];
	fact_table_vect[pos+j+1]=temp;
	/* printf("\n %s",   ft->name);  */

      }
    }
    build_graph_layer(FALSE);

  }

  num_goals_at[p_length] = 0;
  for ( temp = goal_facts; temp; temp = temp->next )
  {
    ft = lookup_from_table( fact_table[p_length], temp->item );
    if(ft){
      goals_at[p_length][num_goals_at[p_length]++] = ft;

      ft->is_goal = 1;

      if ( num_goals_at[p_length] > MAX_GOALS )
      {
	printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );
	exit( 1 );
      }
    }
    else
      {
	insert_incons(pos+j);
      }

  }

  find_local_inc(init_time, p_length);
 /* attenzione i contiene la lungh. del current planning graph */

#ifdef TEST
  printf("\n Stampo fatti setup");
  for(i=0;i<adapt.plan_length;i++){
    printf("\n Fatti level %d: ",i );
    for(temp=fact_table_vect[i]; temp;temp=temp->next )
      printf(" %s ", temp->item);

  }
#endif /*  TEST */
  de_alloc(adapt.plan_length);

}
   
   
void find_fact_goal_op(int time_in, int time_end, int pos)   
{   
  int i, j;   
  vertex_list ft, op;   
  token_list temp;   
  action *temp_act;    
  cond_edge_list i_ce;/* index */

    
  if(time_in<0)  
    time_in=0;
    
  pos=0; 
 
  for ( temp = goal_facts; temp; temp = temp->next )  
    if(( ft = lookup_from_table( fact_table[time_end], temp->item ) ))
      ft->is_used_goal=TRUE;
   
  /* Propago all'indietro i fatti precondiz. di azioni utili aggiornando il campo is_used_goal */
  for(i=time_end-1; i>=time_in && IS_USED==1 ; i--)
    {
      get_next( fact_table[i], INIT );/* initialise */   
      while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL )    
	if( ft->is_goal)
	{  
	  j=0;    
	  ft->is_used_goal=0;
	  for(j=0; j< ft->is_goal; j++)
	    {
	      for ( i_ce = ft->is_goal_for[j]->add_edges; i_ce; i_ce = i_ce->next ) 
	      /* ATTENZ EFF. CONDIZ. NON CONSIDERATI */
		if(i_ce->endpt->is_used_goal)
		  ft->is_used_goal++;
	    	    
	    }
	
	}
    }


  for(i=time_in, j=0; i<=time_end; i++, j++)
  {   
     
    if(goal_table_vect[pos+j])       /*  && i!=time_end */  
    {  
	free_token_list(goal_table_vect[pos+j]);  
	goal_table_vect[pos+j]=NULL;  
    }  
    if( fact_table_vect[pos+j] )   
    {   
	free_token_list(fact_table_vect[pos+j]);   
	fact_table_vect[pos+j]=NULL;   
        
    }   
    
    get_next( fact_table[i], INIT );/* initialise */   
    while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL ) 
    {   
      if( ft->is_goal  && IS_USED==1 && ft->is_used_goal>0)
      {    /*   && ft->is_used_goal>0  */
	temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );   
        temp->item = new_token( strlen( ft->name ) + 1 );    
	strcpy(temp->item, ft->name);   
        temp->next = goal_table_vect[pos+j];   
	goal_table_vect[pos+j]=temp;   
      }   
      /* PROVO A CONSIDERARE COME GOALS SOLO LE PRECONDIZIONI DELLE AZIONI */
      if( ft->is_goal  && IS_USED==4 )
      {
	if(i==time_end || (ft->next_time && ft->next_time->my_noop->is_used))
	{ 
	   temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );   
           temp->item = new_token( strlen( ft->name ) + 1 );    
  	   strcpy(temp->item, ft->name);   
           temp->next = goal_table_vect[pos+j];   
	   goal_table_vect[pos+j]=temp;   
        }
      }

	

#ifdef TEST
      if( ft->is_goal && ft->is_used_goal<=0 )
	printf("\nGoal Removed %s \n", ft->name);
#endif /* TEST */

      if( ft->is_true )
      {   
	temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );     
	temp->item = new_token( strlen( ft->name ) + 1 );      
	strcpy(temp->item, ft->name);   
	temp->next = fact_table_vect[pos+j];   
	fact_table_vect[pos+j]=temp;   
      }    
 
    }   
    
#ifdef TEST
    printf("\n Fatti level %d: ",pos+j ); 
    for(temp=fact_table_vect[pos+j]; temp;temp=temp->next ) 
      printf(" %s ", temp->item); 

    printf("\n Goals level %d: ",pos+j);
    for(temp=goal_table_vect[pos+j]; temp;temp=temp->next )
      printf(" %s ", temp->item);
    fflush(stdout);      
#endif /*  TEST */

    get_next( op_table[i], INIT );/* initialise */   
    while ( ( op = get_next( op_table[i], EXEC ) ) != NULL ) 
    {   
      if( op->is_used  && !op->is_noop)  
	insert_act_vect(op->name, pos+j);
    }   
         
  }    
   
   
  /* Propagazione backward dei goal */
  if(IS_USED==2)
  {
    de_alloc(time_end);

    for(i=0; i<=time_end; i++)
    {
   
      get_next( op_table[i], INIT );/* Rimuovo le azioni  */   
      while ( ( op = get_next( op_table[i], EXEC ) ) != NULL )    
	if( op->is_used)
        {  
           op->is_used=0;
	   op->is_critic=0;
	}

	get_next( fact_table[i], INIT );/* initialise */   
	while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL ) 
	{
	  ft->is_goal = 0;
	  ft->is_true = 0;
	  ft->is_critic=0;
	  ft->is_used_goal=0;
	}
    }
    for ( temp = goal_facts; temp; temp = temp->next )  
      if(( ft = lookup_from_table( fact_table[time_end], temp->item ) ))
	ft->is_used_goal=TRUE;

    for(i=0; i<adapt.plan_length; i++ )                    
      if(act_vect[i])   
	{  
	  for(temp_act=act_vect[i]; temp_act; temp_act=temp_act->psaNext)   
	    {   
	      op = lookup_from_table( op_table[i], temp_act->cStr );   
	      if(op)   
		insert_operators(op, i );  
	      
	    }
	}
 
    for(i=adapt.plan_length; i>0 ; i--)
      {
	if( goal_table_vect[i])
	{
	  free_token_list( goal_table_vect[i]);
	  goal_table_vect[i]=NULL;
	}

	get_next( fact_table[i], INIT );/* initialise */   
	while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL )    
	  if(ft->is_used_goal && ft->is_true==FALSE && ft->my_noop)
	  {
	    /* Tento di propagare con le noop */	
	
	    if(me_op(ft->my_noop, i-1 )==FALSE  )
	      insert_operators( ft->my_noop, i-1); 
	
	  }


	get_next( fact_table[i], INIT );/* initialise */   
	while ( ( ft = get_next( fact_table[i], EXEC ) ) != NULL )    
	  if(ft->is_used_goal)
	  {
	    temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );   
	    temp->item = new_token( strlen( ft->name ) + 1 );    
	    strcpy(temp->item, ft->name);   
	    temp->next = goal_table_vect[i];   
	    goal_table_vect[i]=temp;
	    
	  }
	  
#ifdef TEST
	printf("\n Goal lev %d \n",i);
	print_token_list(  goal_table_vect[i] );
	printf("\n");
#endif 
      
#ifdef TEST

    for ( j=0; j<num_ops_at[i-1]; j++ )
		printf("\n\t Ac: %s num op %d lev%d ", ops_at[i-1][j]->name, num_ops_at[i-1]  , i-1);
#endif /* TEST */
      }
  }
}   


int insert_actions(action *psaFirst, int min_time)
{   
  int diff, time0, result=FALSE;   
  action *az;   
  vertex_list op, ft;   
  token_list it;   
  adapt.num_inc=0;    
      
  expand(min_time);    
  print_facts_and_goals( 0, min_time);
   
  num_goals_at[min_time] = 0;   
  for ( it = goal_facts; it; it = it->next )  
    if(( ft = lookup_from_table( fact_table[min_time], it->item ) ))
    {   
      goals_at[min_time][num_goals_at[min_time]++] = ft;   
      ft->is_goal = 1;   
      ft->is_used_goal=1;
      
      if ( num_goals_at[min_time] > MAX_GOALS ) 
      {   
	printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );   
	exit( 0 );   
      }   
    }       
     
  diff=min_time-adapt.plan_length;   
  time0=psaFirst->iPasso + diff;   
    
  for ( it = initial_facts; it; it = it->next ) 
  {   
    ft = lookup_from_table( fact_table[time0], it->item );   
    if( ft )  
      ft->is_true = 1;   
    
  }   
    
  for(az=psaFirst; az->psaNext ;az=az->psaNext )   
    {   
      op = lookup_from_table( op_table[az->iPasso+diff], az->cStr );   
      if(op)   
	insert_operators(op, az->iPasso+diff );   
      else 
        printf("\nADJ: unable to reuse the action: '%s'\n ", az->cStr);
    }   
     
  /* DA SOSTITUIRE CON temporal projection */    
  insert_noop(time0, min_time);    
      
  find_inc(time0, min_time, FALSE);   
  
  if(adapt.num_inc==0)
     result=TRUE;   
  
#ifdef TEST   
  for(i=0; i<=min_time; i++)   
    {   
      if(incons_vect[i]) printf("\n incons: time %d",i);
      else printf("# ");   
    }   
#endif /* TEST */
   
  adapt.curr_plan_length=min_time;  
  fact_table_vect[0]=initial_facts;;
  find_fact_goal_op(time0, min_time, 0);  
  
  return (result);
}   
  

void insert_operators(vertex_list op, int time)   
{   
   
  edge_list i_e;/* indizes */   
  cond_edge_list i_ce;/* index */   
  int was_used;/* to tell if a fact was already there */   
   
  /* now it's clear that we can put the op into the plan */   
   
  /* mark the op and put it into the global array */   
  was_used = op->is_used++;   
  if ( !was_used && num_ops_at[time] == MAX_GOALS ) 
  {   
    printf( "\n\nipp: increase MAX_GOALS( preset value: %d )",MAX_GOALS );   
    exit( 1 );   
  }   
  if( !was_used ) 
  {   
    ops_at[time][num_ops_at[time]++] = op;   
    num_of_actions_tried++;   
  }   
   
  /* also put the preconditions into goal array at time */   
  for ( i_e = op->precond_edges; i_e; i_e = i_e->next ) 
  {   
     was_used = i_e->endpt->is_goal++;
     i_e->endpt->is_used_goal++;
     if ( ( !was_used && num_goals_at[time] == MAX_GOALS ) ||   
	 i_e->endpt->is_goal > MAX_GOALS ) 
     {   
       printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );   
       exit( 1 );   
     }   
     
    (i_e->endpt->is_goal_for)[i_e->endpt->is_goal-1] = op;   
    if ( !was_used ) 
       goals_at[time][num_goals_at[time]++] = i_e->endpt;   
  }   
   
  /* ... and the effect conditions    
   *   
   * NOTE: only put goals into the array that aren't already there   
   */   
   
  /* now mark the facts that will be true for sure as being so   
   * ( used for speedup ( skip over goals ) and in minimality check )   
   */   
  for ( i_ce = op->add_edges; i_ce; i_ce = i_ce->next )   
    if ( !i_ce->conditions ) i_ce->endpt->is_true++;   
    else /* ATTENZ DA SISTEMARE */   
      for ( i_e =i_ce->conditions; i_e; i_e = i_e->next ) 
      {   
	was_used = i_e->endpt->is_goal++;   
	if ( ( !was_used && num_goals_at[time] == MAX_GOALS ) ||   
	     i_e->endpt->is_goal > MAX_GOALS ) 
	{   
	  printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );    
	  exit( 1 );   
	}   
	(i_e->endpt->is_goal_for)[i_e->endpt->is_goal-1] = op;   
	if ( !was_used ) goals_at[time][num_goals_at[time]++] = i_e->endpt;   
      }   
   
  /* and say we managed */   
   
}   


 
int end_control()
{

  int i, time, result, plan_length;
  action *temp_ac;
  vertex_list op, ft;
  token_list it;

  adapt.num_actions=0;
  adapt.curr_plan_length=0;

  same_as_prev_flag=0;
  expand(adapt.plan_length+20); /* ATT**** */

  adapt.num_inc=0;

  initial_facts=store_initial_facts;
  goal_facts= store_goal_facts;

  de_alloc(adapt.plan_length);
  if(fact_table)
    free_table(fact_table, MAX_PLAN+1);
  fact_table=NULL;
  if(op_table)
    free_table(op_table, MAX_PLAN);
  op_table=NULL;

  plan_length=adapt.plan_length;
  result=build_graph(&plan_length, FALSE);

  get_next( fact_table[0], INIT );
  while ( ( ft = get_next( fact_table[0], EXEC ) ) != NULL )
    ft->is_true = 1;

  for(i=0; i<adapt.plan_length; i++ )
  {
     if(act_vect[i])
     {
	time=act_vect[i]->iPasso;
     	
     	if(adapt.curr_plan_length>=plan_length)
  	  build_graph_layer(FALSE);    	
  	
	for(temp_ac=act_vect[i]; temp_ac; temp_ac=temp_ac->psaNext)
	{
	   op = lookup_from_table( op_table[adapt.curr_plan_length], temp_ac->cStr );
	   if(op)
	   {
	      insert_operators(op, adapt.curr_plan_length );
	      adapt.num_actions++;
#ifdef TEST
	      printf("\n Action %s, level %d",temp_ac->cStr, adapt.curr_plan_length);
#endif /* TEST */
	   }
	   else
	      printf("\nC. There are some errors; debug me please. Action %s, level %d",temp_ac->cStr, adapt.curr_plan_length);
        }

	ins_noop_layer(adapt.curr_plan_length);
	adapt.curr_plan_length++;

     }
     else
        printf("\n Level %d action vect      ", i);
  }

  num_goals_at[adapt.curr_plan_length] = 0;
  for ( it = goal_facts; it; it = it->next )
  {
    ft = lookup_from_table( fact_table[adapt.curr_plan_length], it->item );
    goals_at[adapt.curr_plan_length][num_goals_at[adapt.curr_plan_length]++] = ft;
    ft->is_goal = 1;
    if ( num_goals_at[adapt.curr_plan_length] > MAX_GOALS ) {
      printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );
      exit( 1 );
    }
  }

  for(i=0; i< adapt.curr_plan_length; i++)
  {
     del_inc(i);
     free_act_vect( i );
  }

  find_inc( 0, adapt.curr_plan_length, TRUE );
  find_fact_goal_op(0, adapt.curr_plan_length, 0);

  if(adapt.num_inc)
  {
     printf("\nWARNING: THERE ARE SOME INCONSISTENCES !!!\n\n");
     exit(0);
  }

  adapt.plan_length=adapt.curr_plan_length;

  return(adapt.num_inc);
}
