  
#include "ipp.h" /* defines, data structures, fn prototypes, global variables */  
#include "adapt.h"
 

/* global variables */  
    
/* technical variables */  
char *act_filename;  
int lineno = 1;  
inst_effect_list help_inst_effect_list;/* used for del_ALL_quantifiing */  
BOOLEAN minimize = TRUE;/* has to do with recursive op choice in fn search */  
  
/* globally important variables */  
  
/* domain constants */  
op_list loaded_ops;/* loaded, uninstantiated operators */  
operator_list operators;/* list of instantiated operators */  
token_list initial_facts;/* the initial facts */  
effect_list raw_goal_facts;/* not yet preprocessed goal facts */  
token_list goal_facts;/* the goal facts */  
fact_list orig_constant_list;/* to store all typed objects */  
fact_list orig_initial_facts;/* stores initials as fact_list */  
fact_list negated_preds = NULL;/* needed for NOT preprocessing */   
fact_list types;/* fast access to all objects of one type */  
  
/* have to do with completeness check */  
BOOLEAN same_as_prev_flag = FALSE;/* is set when graph has levelled off */  
int first_full_time;/* stores time when graph has levelled off */  
int previous_Scount = -1;/* previous num hashes at first full time */  
int current_Scount = 0;/* current num hashes at that time */  
  
/* the actual graph: two arrays of hashtables */  
hashtable *fact_table, *op_table;  
  
/* memoizing stuff */  
int simple_hits = 0;/* just for info: */  
int partial_hits = 0;  
int subset_hits = 0;/* different memoize types */  
  
long bytes_graph = 0;  
long bytes_memo = 0;  
  
/* for searching */  
goal_array *goals_at;  
goal_array *ops_at;  
goal_array *critics_at;  
SS_at_type *SS_at;  
int *num_goals_at;  
int *num_ops_at;  
int *num_critics_at;  
int *num_SS_at;  
  
int num_ints = 0;  
  
int num_of_actions_tried = 0;/* info */  
  

/* variables for program timing */

struct tms fix_start, fix_end;
struct tms build_end;


/* setting default values */
 
void initialize_preset_values()
{

  max_nodes = 512;		/* max number of nodes at one level */
  do_subset = TRUE;		/* flag: use subset check in memoizing ? */
  display_info = 1;		/* level of run time info printed */
  do_heuristic = 0;		/* edge ordering strategy : preset none */
  rem_inertia = TRUE;	        /* remove inertia ? */

}


/* setting initial values for adapt */

void initialize_adapt_values()
{

  adapt.plan_length = MAXINT;
  adapt.curr_plan_length = MAXINT;
  adapt.best_plan_length = MAXINT;
  adapt.change;
  adapt.step = STEP;
  adapt.timeout = 1800; 
  adapt.partial_timeout = MAXINT;
  adapt.num_inc = 0;
  adapt.num_actions = 0;
  adapt.best_num_actions = 0;
  adapt.alterna = FWD;               // FWD, BACK
  adapt.iteraz = 0;
  adapt.try = 0;
  
  cut_off = MAXINT;
  iniz_step_cut = MAXINT;
  step_cut = MAXINT;

}


void print_adj_options()
{

  printf("\nUsage of ADJ v1.0\n");
  
  printf("\nOPTIONS   DESCRIPTIONS\n\n");  
  printf("-o <str>    operator file name\n");  
  printf("-f <str>    fact file name\n");  
  printf("-p <str>    path for operator and fact file\n\n");  
  printf("-m <num>    max number of nodes at one level( preset: 512 )\n");  
  printf("-t <num>    minimal planning time( preset: 0 )\n\n");

  printf("-P <str>    ### Input Plan\n"); 
  printf("-T <num>    ### Timeout\n\n");

  printf("-S          don't do partial and complete subset check in memoizing\n");  
  printf("-I          don't remove inertia from planning problem\n\n");  
  printf("-h <num>  heuristic ordering of add edges( preset: 0 )\n");  
  printf("     0    : none\n");  
  printf("     1    : ( sum of pre values / num preconds )\n");  
  printf("     2    : ( product of ( sum of pre values ) )\n\n");  
  printf("-W          write complete graph to text files after planning\n");  
  printf("-n <str>    specify name for graph output files( preset: graph )\n\n");  
  printf("-i <num>  run-time information level( preset: 1 )\n");  
  printf("     0    : only cpu time and num of actions info\n");  
  printf("     1    : plan time info and final plan\n");  
  printf("     2    : info on memoizing hits\n");  
  printf("     3    : display loaded operators and facts\n");  
  printf("     4    : show the instantiable ops\n");  
  printf("     5    : show all the possible instantiations of those\n");  
  printf("     6    : say which inertia are removed\n\n");

  printf("NOTE: if only an operator/fact file <name> is specified,\n");
  printf("      default value for fact/operator file is also <name>.\n");
  printf("      Suffixes '.ops', '.fct' and '.sol' are added automatically. \n\n");
  
}


/* print info on memoizing hits, if wanted */  

void print_info_memoizing()
{

   if (simple_hits!=0 || partial_hits!=0 || subset_hits!=0)
   {
	 
      printf("\nipp: had %7d simple memoizing hits", simple_hits);  
   
      if ( do_subset ) 
      {  
         printf("\nipp  had %7d partial memoizing hits", partial_hits);  
         printf("\nipp: had %7d subset memoizing hits", subset_hits);  
      }
      
   }   
     
   printf("\n\n");  
	  
}


void print_adj_report(float plan_time)
{
   BOOLEAN memory_info=TRUE;
  
   float initial_time;

   printf("\nADJ PLAN STATISTICS \n");   

   printf("\n ADJ: number of total levels: %d ", adapt.plan_length);
   printf("\n ADJ: number of total actions: %d \n", adapt.num_actions);

   initial_time = build_time+build_ad_time;

   printf("\n ADJ: spent  %.2f  seconds initial building graph", initial_time);
   printf("  (initial build  %.2f)  (adapt build  %.2f)", build_time, build_ad_time);
   
   printf("\n ADJ: spent  %.2f  seconds searching for plan", plan_time-initial_time);
   printf("\n ADJ: spent  %.2f  seconds total time \n", plan_time);

   if ( memory_info )
   {
      printf("\n ADJ: memory used for graph: %.1f Kbytes", (float) bytes_graph/1024);
      printf("\n ADJ: memory used for memoizing: %.1f Kbytes \n\n", (float) bytes_memo/1024);
   }

}


void build_file_name(char* full_name, char* file_name,
                     char* file_ext, char* path)
{
   char *sExt;   // This is the extension of a file name

   sExt = file_name;
   sExt += (int) (strlen(file_name) - strlen(file_ext));
     // Get last chars of file name (the extension)
   if ( strcmp ( sExt, file_ext ) )
      strcat( file_name, file_ext );

   strcat( full_name, path );
   strcat( full_name, file_name );
}


/* -------------------- Main ---------------------------- */
  
int main( int argc, char *argv[] )  
  
{  
  
  int exit_status = 0; /* indicates at end if plan was found, used for PURIFY */  
  char ops_file_name[MAX_LENGTH] = "";/* <ops_file_name>.ops */  
  char fct_file_name[MAX_LENGTH] = "";/* <fct_file_name>.fct */  
  char plan_file_name[MAX_LENGTH] = "";/* <plan_file_name>.sol */

  char path[MAX_LENGTH]  = "";/* common path for both files */  
  char ops_file[MAX_LENGTH]  = "";/* resulting name for ops file */  
  char fct_file[MAX_LENGTH]  = "";/* same for fct file */  
  char plan_file[MAX_LENGTH] = "";/* same for plan file */
  
  char save_name[MAX_LENGTH] = "graph";/* preset saving name */  
  
  token_list inertia;/* stores inertia literal names */  
  
  BOOLEAN write_graph = FALSE;/* write graph to output files ? */  
  int min_time = 0;/* specified minimal planning time */  
  int curr_min_time;  
  BOOLEAN reached_goals;/* stores return value of build_graph */  
  int found_plan = FALSE;/* stores return value of search_plan */  
  
  char option;/* command line interpretation */
  
  float plan_time;/* store times */  
  struct tms end;/* helpers */  


  /* global initialization */
  initialize_preset_values();
  initialize_adapt_values();

  /* no options entered: print usage */  
  if (argc == 1) 
  {  
     print_adj_options();
     exit( 1 );  
  }  
 
 
  /* read and interpret command line arguments */  
  while ( --argc && ++argv ) 
  {  
    if ( *argv[0] == '-' && (strlen(*argv) == 2) ) 
    {  
      option = *++argv[0];  
      switch ( option ) {/* first check the flags */  
      case 'S':  
        do_subset = FALSE;  
        break;  
      case 'W':  
        write_graph = TRUE;  
        break;   
      case 'I':  
        rem_inertia = FALSE;  
        break;  

      default:  
        if ( --argc && ++argv ) 
	{ /* option specified */  
	  switch ( option ) 
	  { /* examine the option letter */  
	  case 'o':  
	    strncpy(ops_file_name, *argv, MAX_LENGTH);  
	    break;  
	  case 'f':  
	    strncpy(fct_file_name, *argv, MAX_LENGTH);  
	    break;  
          case 't':  
            sscanf(*argv, "%d", &min_time);  
            break;  
          case 'T':  
	    sscanf( *argv,"%d",&adapt.timeout );  
	    break;  
          case 'm':  
            sscanf(*argv, "%d", &max_nodes);  
            break;  
          case 'i':  
            sscanf(*argv, "%d", &display_info);  
            break;  
          case 'h':  
            sscanf(*argv, "%d", &do_heuristic);  
            break;  
          case 'p':  
	    strncpy(path, *argv, MAX_LENGTH);  
            break;  
          case 'n':  
	    strncpy(save_name, *argv, MAX_LENGTH);  
            break;  
          case 'P':  
            strncpy(plan_file_name, *argv, MAX_LENGTH);    
            break;  

	  default:  
 	    printf( "\nipp: unknown option: %c entered\n\n", option );  
	    exit( 1 );  
	    break;  
	  }  
        } 
	else 
	{/* something wrong with the option */  
	  printf("\nipp: unknown option\n\n");  
	  exit( 1 );  
        }  
      }  
    } 
    else 
    {/* option has more than one letter or no '-' prefix */  
      printf("\nipp: wrong usage\n\n");  
      exit( 1 );  
    }  
  }  
  
  /* unknown heuristc entered */  
  if (  do_heuristic > 2 ) 
  {  
    printf( "\nipp: unknown heuristic entered\n\n" );  
    exit( 1 );  
  }  
  
  
  
 /* too large max_nodes as argument of -m specified (Jana 17/9) */   
  if (  max_nodes > MAX_MAX_NODES ) 
  {  
    printf( "\nipp: -m %d exceeds MAX_MAX_NODES bound of %d\n",   
               max_nodes,  MAX_MAX_NODES );    
    exit( 1 );  
  }  
  
  /* no input name at all entered */  
  if ( !ops_file_name[0] && !fct_file_name[0] )
  {  
    printf( "\nipp: at least one inputfile name needed\n\n" );  
    exit( 1 );  
  }  
  
  /* only one entered ? => copy from other */  
  if ( !ops_file_name[0] ) strncpy( ops_file_name, fct_file_name, MAX_LENGTH);  
  if ( !fct_file_name[0] ) strncpy( fct_file_name, ops_file_name, MAX_LENGTH);  
  
/* add type suffixes: extension for input files and path info */
  build_file_name(ops_file, ops_file_name, ".ops", path);
  build_file_name(fct_file, fct_file_name,".fct", path);
  build_file_name(plan_file, plan_file_name,".sol", path);

  load_fct_file( fct_file );  
  load_ops_file( ops_file );  

  
  /* preprocessing: enable to use negated facts */
  preprocess_not();  /* globals changed: negated_preds, initial_facts */
  
  /* convert initials to graph storage */
  initial_facts = token_list_from_fact_list( orig_initial_facts );
  
  /* preprocessing: remove quantifiers, remove eq() predicates */
  goal_facts = dequantify_goals( raw_goal_facts, &loaded_ops );
  
  /* display info on loaded operators and fact, if wanted */
  if ( display_info > 2 )
  {
     printf( "\nloaded operators are:\n" );
     print_ops( loaded_ops );
     printf( "\nloaded facts are:\n" );
     print_fct( orig_constant_list, initial_facts, goal_facts );
     printf( "\n" );
  }
  
  /* remove operators that cannot be instantiated with given facts */
  remove_uninstantiable_ops( &loaded_ops, orig_constant_list );
  
  /* display info on not removed operators, if wanted */
  if ( display_info > 3 )
  {
     printf( "\nnot removed operators are:\n" );
     print_ops( loaded_ops );
     printf( "\n" );
  }
  else
     printf( "\n" );

  /* begin timing */
  times( &start_time );
  
  /* find out which literals are inertia */
  inertia = get_inertia( orig_initial_facts, loaded_ops );
  
  /* calculate all possible instantiations
   * and store them in global list operators
   */
  instantiate( loaded_ops, orig_initial_facts, inertia, orig_constant_list );
  
  /* find and remove all facts that are never deleted from the
   * global op and fact lists
   */
  if ( rem_inertia )
     handle_inertia( inertia );
  
  /* this one goes through all instantiated operators, looking
   * for effects with identical conditions that might occur due to
   * removal of inertia or eq - conditions
   *
   * is defined in inertia.c, which doesn't make sense but saves us from
   * yet another file of c code
   */
  merge_identical_effects();

  /* display info on instantiated operators, if wanted */
  if ( display_info > 4 )
  {
     printf( "\n Original instantiated operators are:\n\n" );
     print_operators( operators );
     printf("\n" );
  }


/* ------------ Start of Adapt ------------- */

  printf("\nSTARTING ADJ ... \n");
     
  load_plan(plan_file_name);  
  if( min_time < psaLast->iPasso)   
     adapt.plan_length=min_time= psaLast->iPasso+1;  

/* initial facts and goals storing */      
   store_initial_facts = copy_token_list( initial_facts );
   store_goal_facts = copy_token_list( goal_facts );
  
/* build graph until non-exclusive goals are reached */  
   reached_goals = build_graph( &min_time, TRUE );  
  
/* get building time */  
   times( &end );  
   
   build_time = ( float ) ( ( end.tms_utime - start_time.tms_utime +  
	              	      end.tms_stime - start_time.tms_stime  ) / 100.0 ) ;  
  
   adapt.curr_plan_length=min_time;
   found_plan=insert_actions(psaFirst, min_time);  

   printf("\nSEARCHING FOR SOLUTIONS ... \n");
   times(&search_start);
      
   if ( !reached_goals ) 
   {  
      if ( min_time < MAX_PLAN ) 
      {  
	 /* graph has leveled off and goals are still not reached */  
	 printf("\nipp: problem unsolvable: can't reach non exclusive goals\n\n");  
      } 
      else 
      {  
         /* we didn't even have time to find the goals */  
	 printf( "\nipp: MAX_PLAN too small( preset value: %d )\n\n", MAX_PLAN );  
      }  
   }

   else 
   {  
      if ( min_time ) 
      {  
      
         curr_min_time = choose_inc();
         
         if( curr_min_time < 0 )
         {
            found_plan = TRUE;
            printf("\nADJ: the stored plan solve the planning problem !\n");
         }
         
         else
         {        
            /* found an inconsistence at level higher than 0; start searching for a plan */ 
	      
            while ( curr_min_time >= 0 &&  adapt.num_inc > 0 &&  adapt.plan_length < MAX_PLAN )
            {  
  	       found_plan = adapt_plan(curr_min_time);
               curr_min_time = choose_inc();
          
               if ( !found_plan || found_plan == TIMEOUT )
	          break; 	

            }
         }


	 if ( found_plan == TIMEOUT )	
	    printf("\nADJ: timeout reached, please increase time limit! \n\n");

         else if ( found_plan )
	 {
	    /* calculate and print running time */

	    times( &end );
	    end_control();

	    adapt.best_num_actions = adapt.num_actions;
	    adapt.best_plan_length = adapt.curr_plan_length;

	    plan_time = ( float ) ( ( end.tms_utime - start_time.tms_utime +
        	                      end.tms_stime - start_time.tms_stime  ) / 100.0 ) ;

	    printf("\n\nFIRST ADJUSTED PLAN \n");
	    print_plan( adapt.curr_plan_length );
	    print_adj_report(plan_time);

	    // this is only for testing
	    #ifdef TEST_ADJ
	       printf("\nAABBCC\t ADJ: spent  \t %.2f  seconds total time; num. of levels  \t %d num. of actions \t %d\n",
        	      plan_time, adapt.plan_length,  adapt.num_actions );
	    #endif /* TEST_ADJ */

	    store_adapted_plan(adapt.curr_plan_length, fct_file_name, plan_time);
	    fflush(stdout);

	    printf("\nSTARTING OPTIMIZATION PROCESS ... \n");
	    resize_plan(fct_file_name);
	
	 }

         else
         {

            if ( min_time < MAX_PLAN )
            {  
               /* or there is no plan for the problem, */  
               printf( "\nipp: problem proved unsolvable\n\n" );  
          
               if ( display_info ) 
 	          print_info_memoizing();	     	  
            }
         
            else 
            {  
               /* or we didn't give the planner enough time */  
               printf("\nipp: MAX_PLAN too small( preset value: %d )\n\n",MAX_PLAN);  
            }  
         }        
      } 
      
      else
      {
         /* corresponds to if ( min_time ) */
         printf( "\nipp: goals are contained in initial state\n\n" );  
    }

    /* create the graph files, if required */
    if ( write_graph )
    {
       printf("\nipp: writing graph\n\n");
       if ( !SaveGraph( save_name, min_time ) )
	  printf("\nipp: error writing graph\n\n");

    }
  }
       
  if (found_plan)
     exit_status = 0;
  else
     exit_status = 1;
  
  printf("\nEXITING FROM ADJ ... \n\n\n");
  exit( exit_status );
  
}  
