%{
#include <stdio.h>
#include <string.h> 
#include "ipp.h"

int act_err;
char *act_err_par;
fact_list act_var, act_const;
effect_list q_goal_list;
inst_effect_list inst_goal_list;
token_list token_ini_facts;
token_list act_goal;
fact_list act_pardec;
fact_list new_neg_pred;

#ifndef SCAN_ERR
#define SCAN_ERR
#define OPNAME_EXPECTED       0
#define PARTOK_EXPECTED       1
#define PARDEC_EXPECTED       2
#define PRETOK_EXPECTED       3
#define PRECONDS_EXPECTED     4
#define EFFTOK_EXPECTED       5
#define EFFECTS_EXPECTED      6
#define TYPENAME_UNKNOWN      7
#endif
static char *errmsg[] = {
  "operator name expected",
  "\':v\' expected",
  "parameter declaration expected",
  "\':p\' expected",
  "list of preconditions expexted",
  "\':e\' expected",
  "list of add & del effects expexted",
  "unknown type '%s' (check for typing errors...)",

  NULL
};

void opserr( int errno, char *par );

%}



%start opfile

%union {
  char string[256];
  fact_list fact_list;
  op_list op_list;
  token_list token_list;
  effect_list effect_list;
}

%type <op_list> operator
%type <fact_list> factlist
%type <fact_list> pardec
%type <fact_list> varlist
%type <token_list> args
%type <effect_list> effects
%type <fact_list> addlist
%type <fact_list> dellist
%type <effect_list> cont



%token <string> VARNAME 
%token <string> COMMENT
%token <string> CONSTNAME
%token <string> PREDNAME
%token <string> TYPENAME
%token IMPL
%token PAR
%token PRE
%token EFF
%token ALL
%token DEL
%token ADD
%token NOT

%%


/* numbers in comments are only useful for the programmer
   and reader of the yacc source code. Each terminal or
   non-terminal symbol and (which is not nice) each
   program section in {brackets} has a number and may
   be reached by the $ operator - the numbers only
   help prevent counting all the time and produce
   ugly fat spots on the monitior */


opfile          : operator { loaded_ops = $1; }
                ;

operator        : /* empty */
                    { $$ = NULL; } 
                | { opserr( OPNAME_EXPECTED, NULL ); } CONSTNAME       /* 1 2 */
		  { opserr( PARTOK_EXPECTED, NULL ); } PAR             /* 3 4 */
		  { opserr( PARDEC_EXPECTED, NULL ); } pardec          /* 5 6 */
		  { opserr( PRETOK_EXPECTED, NULL ); } PRE             /* 7 8 */
		  { opserr( PRECONDS_EXPECTED, NULL ); } factlist      /* 9 10 */
		  { opserr( EFFTOK_EXPECTED, NULL ); } EFF             /* 11 12 */
                  { opserr( EFFECTS_EXPECTED, NULL ); } effects        /* 13 14 */
		  { opserr( OPNAME_EXPECTED, NULL ); } operator        /* 15 16 */
                    {
		      $$ = ( op_list ) malloc( sizeof( op_list_elt ) );
		      $$->name = ( token ) calloc( 1 + strlen( $2 ), sizeof( char ) );
		      strcpy( $$->name, $2 );
		      $$->params = $6;
		      $$->preconds = $10;
		      $$->effects = $14;
		      $$->next = $16;
		    }
                ;


pardec          : /* empty */
                    { $$ = NULL; }
                | varlist CONSTNAME pardec
                    {
		      $$ = $1;
		      for( act_var=$$; act_var; act_var=act_var->next )
			{ /* adjust types (2nd element of token_list) for varlist */
			  act_var->item->next = 
			    ( token_list ) malloc( sizeof( token_list_elt ) );
			  act_var->item->next->item = 
			    ( token ) calloc( 1 + strlen( $2 ), sizeof( char ) );
			  strcpy( act_var->item->next->item, $2 );
			  act_var->item->next->next = NULL;
			  if ( !act_var->next ) {
			    act_var->next = $3; 
			    break;
			  }
			}
		    }
		;
	  
varlist         : /* empty */ 
                    { $$ = NULL; }
	        | VARNAME varlist
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->next = $2;
		    }
		;
      
factlist       : /* empty */ 
                    { $$ = NULL; }
                | PREDNAME '(' args ')' factlist
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item->item, $1 );
		      $$->item->next = $3;
		      $$->next = $5;
             	    }
                | NOT PREDNAME '(' args ')' factlist
                    {
		      $$ = ( fact_list ) malloc( sizeof( fact_list_elt ) );
		      $$->item = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item->item = 
			( token ) calloc( 2 + strlen( $2 ), sizeof( char ) );
		      strcpy( $$->item->item, "!" );
		      strcat( $$->item->item, $2 );
		      $$->item->next = $4;
		      $$->next = $6;
		    }
                  ; 

args            : /* empty */
                    { $$ = NULL; }
                | CONSTNAME args
                    {
		      $$ = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item = ( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item, $1 );
		      $$->next = $2;
		    }
                | VARNAME args
                    {
		      $$ = ( token_list ) malloc( sizeof( token_list_elt ) );
		      $$->item = ( token ) calloc( 1 + strlen( $1 ), sizeof( char ) );
		      strcpy( $$->item, $1 );
		      $$->next = $2;
		    }
		;      

/* deckt mehr als die bisher zulaessige syntax ab, zb muss zwischen
   vorbed. und effects kein => zeichen mehr stehen. der parser kann
   trotzdem vorbedingungen von effekten unterscheiden */
                
effects         : ALL pardec factlist impl addlist dellist cont
                    {
		      $$ = ( effect_list ) malloc( sizeof( effect_list_elt ) );
		      $$->quantified_variables = $2;
		      $$->conditions = $3;
		      $$->add_effects = $5;
		      $$->del_effects = $6;
		      $$->next = $7; /* NULL or another effect_list */
		    }
                | factlist impl addlist dellist cont
                    {
		      $$ = ( effect_list ) malloc( sizeof( effect_list_elt ) );
		      $$->quantified_variables = NULL;
		      $$->conditions = $1;
		      $$->add_effects = $3;
		      $$->del_effects = $4;
		      $$->next = $5; /* NULL or another effect_list */
		    }
                ;

addlist         : /* empty */
                    { $$ = NULL; }
                | ADD factlist 
                    { $$ = $2; }
                ;

dellist         : /* empty */
                    { $$ = NULL; }
                | DEL factlist 
                    { $$ = $2; }
                ;

/* another effect or the end of an operator */
cont            : '.' { $$ = NULL; } 
                | ';' effects { $$ = $2; }

impl            :  /* empty */
                | IMPL 
                ;



%%

#include "lex.ops.c"

void opserr( int errno, char *par )
{
  act_err = errno;
  act_err_par = par;
}
  
int yyerror( char *msg )
{
  printf("%s: syntax error in line %d, '%s':\n", act_filename, lineno, yytext );
  if ( act_err_par )
    {
      printf( errmsg[act_err], act_err_par );
      printf( "\n" );
    }
  else
    printf("%s\n", errmsg[act_err] );
  exit( 1 );
}


void load_ops_file( char *filename )
{
  FILE *fp;/* pointer to input files */

  /* open operator file */
  if( ( fp = fopen( filename, "r" ) ) == NULL ) {
    printf( "\nipp: can't find operator file: %s\n\n", filename );
    exit( 1 );
  }
  act_filename = filename;
  lineno = 1; 
  yyin = fp;
  yyparse();
  fclose( fp );/* and close file again */
}


void print_ops( op_list op )
{
  
  op_list    i_op;
  effect_list  i_list;

  for ( i_op = op; i_op; i_op = i_op->next ) {
    
    printf( "\n%s \n", i_op->name );

    printf( "v: " );
    print_factlist( i_op->params );

    printf( "\np: ");
    print_factlist( i_op->preconds );

    printf( "\ne: ");
    for ( i_list = i_op->effects; i_list; i_list = i_list->next ) {
      if ( i_list->quantified_variables ) {
        printf( "ALL " );
        print_factlist( i_list->quantified_variables );
      }
      if ( i_list->conditions ) {
        print_factlist( i_list->conditions );
        printf( "=> " ); 
      }
      if ( i_list->add_effects ) {
        printf( "ADD " );
        print_factlist( i_list->add_effects );
        if ( i_list->del_effects ) {
          printf( "DEL " );
          print_factlist( i_list->del_effects );
          printf( ";\n" );
        } else {
          printf( ";\n" );
        }
      } else {
        printf( "DEL " );
        print_factlist( i_list->del_effects );
        printf( ";\n" );
      }
    }
  }
}

