
#include "adapt.h"
#include "ipp.h"


// The following functions are for action vector managing// Insert an action, if not present

BOOLEAN find_act_vect(token name, int level)
{
   action *temp;

   for(temp=act_vect[level]; temp; temp=temp->psaNext)
   {
      if (strcmp(temp->cStr, name) == SAME)
    	return (TRUE);
   }      	

   return (FALSE);
}


// Insert an action, if not present

void insert_act_vect(token name, int level)
{    
   action *temp_act ;
   int position=level;

   if ( !find_act_vect(name, level) )
   {
      temp_act =  ( action * ) calloc( 1, sizeof( action ) );
      strcpy(temp_act->cStr, name);
      temp_act->iPasso=position*100;
      temp_act->psaNext= act_vect[level];
      act_vect[level]=temp_act;
   }

}   
   

// Free an entire level of action vector
   
void free_act_vect(int level)   
{    
   action *temp_act, *temp;    
   
   for(temp_act=act_vect[level]; temp_act; )   
   {   
      temp = temp_act;   
      temp_act = temp->psaNext ;   
      free(temp);   
   }   
  
   act_vect[level]=NULL;      
}   


// All levels from "level" are moved one step up, when we want to insert a new (longer) plan

void up_act_vect(int level)
{
   int i;

   for ( i= adapt.plan_length;i>=level; i--)
   {
      act_vect[i+1]= act_vect[i];       
      goal_table_vect[i+1]=goal_table_vect[i];
      correct_vect[i+1]=correct_vect[i];
   }
   
   act_vect[level]=NULL;
   goal_table_vect[level]=NULL;
   adapt.plan_length++;
}


// All levels from "level" are moved one step down, when we want to insert a new (shorter) plan

void down_act_vect(int level)
{
   int i;

   if(act_vect[level])
      free_act_vect(level);

   for (i=level; i<=adapt.plan_length; i++)
   {
      act_vect[i]=act_vect[i+1];
      goal_table_vect[i]=goal_table_vect[i+1];
      correct_vect[i]=correct_vect[i+1];
   }
   
   act_vect[i]=NULL;
   goal_table_vect[i]=NULL;
   adapt.plan_length--;
}


void insert_noop(int init, int end)   
{   
  int i;
  
  for(i=init; i<end ; i++)
    ins_noop_layer(i);

}   
  
   
void ins_noop_layer(int time)  
{  
  
  vertex_list ft;  
  
  get_next( fact_table[time], INIT );/* initialise */   
  while ( ( ft = get_next( fact_table[time], EXEC ) ) != NULL ) 
  {   
    if(  ft->is_true && !me_noop(ft->next_time->my_noop, time) )
    {   
      insert_operators(ft->next_time->my_noop, time);   
     
    }   
  }   
}  


BOOLEAN me_op(vertex_list op, int time)
{

  int i;
    vertex_list o;

   for ( i=0; i<HSIZE; i++ )
      for ( o = op_table[time][i]; o; o = o->next )
  	if(o->is_used && ARE_MUTEX( o, op ) )
	  return TRUE;

  return FALSE;

}


BOOLEAN me_noop(vertex_list op, int time)
{

  int i;
  vertex_list o;

  for ( i=0; i<HSIZE; i++ )
     for ( o = op_table[time][i]; o; o = o->next )
     	if(o->is_used && !o->is_noop && ARE_MUTEX( o,op))
	   return TRUE;

  return FALSE;

}



BOOLEAN me_fct(vertex_list fct, int time)
{
  vertex_list ft;
  int j;

  for (j = 0; j<HSIZE; j++ )
     for ( ft = fact_table[time][j]; ft; ft = ft->next )
        if( ft->is_true && ft!=fct  && ARE_MUTEX( ft, fct ) ) 	
           return TRUE;

  return FALSE;
}


void insert_incons(int time)
{
  incons_vect[time]++;

  if( incons_vect[time]==1 )
    adapt.num_inc++;

}


void del_inc(int time)
{
  if( incons_vect[time]==0 )
     return;
  incons_vect[time]=0;
  adapt.num_inc--;

}


int choose_inc()
{
  int i;

  if(adapt.num_inc==0)
    return -1;

  for(i=0; i<MAX_PLAN ; i++)
    if( incons_vect[i] )
      return i;

  return -1;
}


void de_alloc(int max_time)
{

  int i;
  
  choose_critics(max_time, CLEAR);  
  
  for ( i=0; i < max_time; i++ )
  {
     num_goals_at[i]=0;
     num_ops_at[i]=0;
     num_SS_at[i]=0;
     num_critics_at[i]=0;
  }

/*  choose_critics(max_time, CLEAR); */
}


void free_table(hashtable *t, int max_size)

{
  int i, j;
  vertex_list v, v1;

  for(i=0;i<max_size;i++)
    if(t[i]!=NULL)
      for(j=0; j<HSIZE; j++)
      {
	for( v = t[i][j]; v!=NULL; ) 
	{
	  v1=v;
	  v = v->next;
	  free_vertex(v1);
	}
	t[i][j]=NULL;
      }
  free(t);
}


void free_edge_list(edge_list e)
{
  edge_list e_l;
  for(; e ;)
    {
      e_l=e;
      e=e->next;
      free(e_l); 
    }

}


void free_cond_edge_list(cond_edge_list ce)
{
  cond_edge_list ce_l;
  for(; ce ;)
    {
      ce_l=ce;
      ce=ce->next;
      if(ce_l->conditions) 
	free_edge_list(ce_l->conditions);
   
      free(ce_l); 
    }

}


void free_delete_list(delete_list de)
{
  delete_list de_l;
  for(; de ;)
    {
      de_l=de;
      de=de->next;
      if(de_l->conditions)
	free_edge_list(de_l->conditions);
    
      free(de_l); 
    }

}


void free_vertex(vertex_list v)
{
  
  if(v->name)
    free(v->name);
  /*    else return; */
  v->name=NULL;
  if( v->precond_edges )
    free_edge_list( v->precond_edges );
  v->precond_edges=NULL;
  if( v->add_edges)
    free_cond_edge_list( v->add_edges);
  v->add_edges=NULL;
  if( v->del_list)
    free_delete_list( v->del_list);
  v->del_list=NULL;
  if(v->memo_start)
    free_sub_tree(v->memo_start);
  v->memo_start=NULL;

  free(v);

}


void print_partial_plan(int time)
{

  int i,j;      /* index for time */
  BOOLEAN first;/* for better printing: is it the first one this level? */

  vertex_list op;

  printf("\nADJ FOUND PLAN AS FOLLOWS\n\n");

  /* for all non empty operator levels */
  for ( i=0; i<time; i++ )
  {
     printf("time step %3d: ", i);
     first = TRUE;

     /* for all operators in plan */
     for ( j=0; j<num_ops_at[i]; j++ )
     {
        op = ops_at[i][j];

        if ( op->is_noop ) continue;/* don't print noops */

        if ( first )
        {
	   printf("\t%s\n", op->name );
	   first = FALSE;
        }
        else
      	   printf("\t\t%s\n", op->name );
     }
  }

  printf( "\n" );

}


void print_plan(int time)
{

  int i;          /* index for time */
  BOOLEAN first;  /* for better printing: is it the first one this level? */

  action *temp_act;

  printf("\nADJ FOUND PLAN AS FOLLOWS\n\n");

  /* for all non empty operator levels */
  for ( i=0; i<time; i++ )
  {
     printf("time step %3d: ", i);
     first = TRUE;

     /* for all operators in plan */
     for(temp_act = act_vect[i]; temp_act; temp_act = temp_act->psaNext)
     {
        if ( first )
        {
           printf("\t%s\n", temp_act->cStr );
	   first = FALSE;
        }
        else
	   printf("\t\t%s\n", temp_act->cStr );
     }
  }

  printf( "\n" );

}



void print_facts_and_goals(int curr_init_time, int curr_max_time)
{

   printf("\nINITIAL FACTS   (time %d) \n", curr_init_time);
   print_list(initial_facts);  
   
   printf("\nGOALS           (time %d) \n", curr_max_time);
   print_list(goal_facts);
   printf("\n");
}


// Printing of a token list : for video output

void print_list(token_list list)
{

  token_list i_token;

  for ( i_token = list; i_token; i_token = i_token->next )
    printf( "\n - %s ", i_token->item );
    
  printf("\n");
     
}

